
.widget-topk-vis {
  width: calc(100%);
  height: 100%;
  padding: 10px;
  color: #111;
}
.sampleActivationBar {
  width: 95%; 
  border: 1px solid lightgray;
  height: 14px;
  position: relative;
  margin: 1px;
  overflow: hidden;
}
.sampleActivationBarForeground {
  height: 12px;
  opacity: 0.7;
}
.sampleActivationBarForeground:hover {
  opacity: 1;
}
.sampleActivationBarLabel {
  font-size: 10px;
  position: absolute;
  top: -2px;
  left: 4px;
  display: flex;
  flex-direction: row;
  align-items: center;
  justify-content: space-between;
  pointer-events: none;
  width: 95%;
  overflow: hidden;
}

